#ifndef HitsSoA_h
#define HitsSoA_h

#include <algorithm>
#include <cstdint>
#include <vector>

/**
 * @class HitsSoA
 * @brief Represents Hits stored in a Structure of Arrays (SoA) format.
 *
 * The HitsSoA data structure provides a convenient way to store and access Hits data in a
 * Structure of Arrays (SoA) format. It allows for efficient memory access and cache utilization
 * when working with large collections of Hits.
 */
struct HitsSoA {

  std::vector<float> energy;
  std::vector<float> x;
  std::vector<float> y;
  std::vector<float> z;
  std::vector<float> time;
  std::vector<float> fraction; // for simhits
  std::vector<uint32_t> id;
  std::vector<int> clusterIndex; // for simhits

  void resize(unsigned int size) {
    energy.resize(size);
    x.resize(size);
    y.resize(size);
    z.resize(size);
    time.resize(size);
    fraction.resize(size);
    id.resize(size);
    clusterIndex.resize(size);
  }

  void clear() {
    energy.clear();
    x.clear();
    y.clear();
    z.clear();
    time.clear();
    fraction.clear();
    id.clear();
    clusterIndex.clear();
  }

  // this overload extends the vectors in the HitsSoA
  // and is used when merging SimClusters
  HitsSoA operator+=(HitsSoA &toAdd) {
    unsigned int nHitsToAdd = toAdd.id.size();
    for (unsigned int i = 0; i < nHitsToAdd; ++i) {
      std::vector<uint32_t>::iterator findId =
          std::find(this->id.begin(), this->id.end(), toAdd.id[i]);
      if (findId == this->id.end()) {
        this->energy.push_back(toAdd.energy[i]);
        this->x.push_back(toAdd.x[i]);
        this->y.push_back(toAdd.y[i]);
        this->z.push_back(toAdd.z[i]);
        this->time.push_back(toAdd.time[i]);
        this->fraction.push_back(toAdd.fraction[i]);
        this->id.push_back(toAdd.id[i]);
      } else {
        auto hitIndex = std::distance(id.begin(), findId);
        this->energy[hitIndex] += toAdd.energy[i];
        this->fraction[hitIndex] += toAdd.fraction[i];
      }
    }
    return *this;
  }
};

#endif
