#ifndef MergeVertex_h
#define MergeVertex_h

#include "SimClusterMergeWrapper.h"
#include <string>
#include <vector>
#include <memory>


template <class T, class Compare>
bool _isInSorted(const std::vector<T> &a, const T &cmp, Compare __c,
                 size_t &start) {
  for (size_t i = start; i < a.size(); ++i) {
    if (__c(a.at(i), cmp)) {
      start = i;
      continue;
    }
    if (a.at(i) == cmp) {
      start = i + 1;
      return true;
    }
  }
  return false;
}

template <class T> bool isInSorted(const std::vector<T> &a, const T &cmp) {
  size_t start = 0;
  return _isInSorted(a, cmp, std::less<T>(), start);
}

template <class T, class Compare>
bool _isInSorted(const std::vector<T> &a, const T &cmp, Compare __c) {
  size_t start = 0;
  return _isInSorted(a, cmp, __c, start);
}

template <class T>
bool isInSorted(const std::vector<T> &a, const T &cmp, size_t &start) {
  return _isInSorted(a, cmp, std::less<T>(), start);
}

template <class T, class Compare>
bool isInSorted(const std::vector<T> &a, const T &cmp, size_t &start,
                Compare __c) {
  return _isInSorted(a, cmp, __c, start);
}

template <class T, class Compare>
void _intersectSorted(const std::vector<T> &a, const std::vector<T> b,
                      std::vector<size_t> &int_a, std::vector<size_t> &int_b,
                      Compare __c, std::vector<size_t> &u_a,
                      std::vector<size_t> &u_b, bool return_unique) {
  int_a.clear();
  int_b.clear();
  if (return_unique) {
    u_a.clear();
    u_b.clear();
  }
  if (a.size() < 1 || b.size() < 1)
    return;
  size_t a_idx = 0, b_idx = 0;
  while (true) {
    if (a_idx >= a.size())
      break;
    const T &cmp = a.at(a_idx);
    while (b_idx < b.size() && __c(b.at(b_idx), cmp)) {
      b_idx++;
    }
    if (b_idx >= b.size() && !return_unique) {
      break;
    } else if (b_idx < b.size() && cmp == b.at(b_idx)) {
      int_a.push_back(a_idx);
      int_b.push_back(b_idx);
    } else if (return_unique) {
      u_a.push_back(a_idx);
    }
    a_idx++;
  }
  if (!return_unique)
    return;

  size_t int_b_count = 0;
  for (size_t i = 0; i < b.size(); ++i) {
    if (!isInSorted(int_b, i, int_b_count)) {
      u_b.push_back(i);
    }
  }
}

template <class T>
void intersectSorted(const std::vector<T> &a, const std::vector<T> b,
                     std::vector<size_t> &int_a, std::vector<size_t> &int_b) {
  std::vector<size_t> u_a, u_b;
  _intersectSorted(a, b, int_a, int_b, std::less<T>(), u_a, u_b, false);
}

template <class T, class Compare>
void intersectSorted(const std::vector<T> &a, const std::vector<T> b,
                     std::vector<size_t> &int_a, std::vector<size_t> &int_b,
                     Compare __c) {
  std::vector<size_t> u_a, u_b;
  _intersectSorted(a, b, int_a, int_b, __c, u_a, u_b, false);
}

template <class T>
void intersectSorted(const std::vector<T> &a, const std::vector<T> b,
                     std::vector<size_t> &int_a, std::vector<size_t> &int_b,
                     std::vector<size_t> &u_a, std::vector<size_t> &u_b) {
  _intersectSorted(a, b, int_a, int_b, std::less<T>(), u_a, u_b, true);
}

template <class T, class Compare>
void intersectSorted(const std::vector<T> &a, const std::vector<T> b,
                     std::vector<size_t> &int_a, std::vector<size_t> &int_b,
                     std::vector<size_t> &u_a, std::vector<size_t> &u_b,
                     Compare __c) {
  _intersectSorted(a, b, int_a, int_b, __c, u_a, u_b, true);
}

template <class T, class Compare>
void _insertSortedInSorted(const std::vector<T> &toinsert,
                           std::vector<T> &target, Compare __c) {
  auto old = target;
  target.clear();

  size_t inscounter = 0;
  for (size_t i = 0; i < old.size(); ++i) {
    bool newfirst = false;
    if (inscounter < toinsert.size())
      newfirst = __c(toinsert.at(inscounter), old.at(i));
    if (newfirst) {
      target.push_back(toinsert.at(inscounter));
      inscounter++;
      i--;
    } else {
      target.push_back(old.at(i));
    }
  }

  while (inscounter < toinsert.size()) {
    target.push_back(toinsert.at(inscounter));
    inscounter++;
  }
}

template <class T, class Compare>
void insertSortedInSorted(const std::vector<T> &toinsert,
                          std::vector<T> &target, Compare __c) {
  _insertSortedInSorted(toinsert, target, __c);
}

template <class T>
void insertSortedInSorted(const std::vector<T> &toinsert,
                          std::vector<T> &target) {
  _insertSortedInSorted(toinsert, target, std::less<T>());
}


/**
 * @class MergeGraph
 * @brief Represents a graph for merging vertices.
 * 
 * The MergeGraph class is a template class that represents a graph for merging vertices.
 * It stores a collection of vertices and edges between the vertices.
 * The class provides methods to set the vertices, add a vertex, create edges between the vertices, merge the vertices, and print the edges.
 * 
 * @tparam T The type of the vertices in the graph.
 */

template <class T>
class MergeVertex {

public:
  MergeVertex() : isLow_(true) {}
  MergeVertex(T* t) : isLow_(true) {
    obj_.push_back(t);
    name_ = std::to_string(n_mergeVertex);
    n_mergeVertex++;
  }

  virtual ~MergeVertex() {}

  const std::string &name() const { return name_; }

  bool isLow() const { return isLow_; }
  bool isHigh() const { return !isLow_; }

  void setLow(bool islow) { isLow_ = islow; }

  void absorb(MergeVertex<T>* rhs);

  std::vector<T*> objects() const { return obj_; }

  bool valid() const { return obj_.size() > 0; }

  virtual void updateIsLow() {}

  bool shouldConnectBase(const MergeVertex* rhs) const;
  virtual bool shouldConnect(const MergeVertex* rhs) const {
    return shouldConnectBase(rhs);
  }

  void clear() { obj_.clear(); }
protected:
  std::string name_;
  std::vector<T*> obj_;
  int n_mergeVertex;
  bool isLow_;
};

template <class T> void MergeVertex<T>::absorb(MergeVertex<T>* rhs) {
  if (this == rhs)
    return;
  obj_.insert(obj_.end(), rhs->obj_.begin(), rhs->obj_.end());
  rhs->clear();
  updateIsLow();
  name_ = name_ + ", " + rhs->name_;
  rhs->name_ = "(" + rhs->name_ + ")";
}

template <class T>
bool MergeVertex<T>::shouldConnectBase(const MergeVertex *rhs) const {
  if (this == rhs || !rhs->valid() || !valid() || rhs->isHigh())
    return false;
  return true;
}

#endif
