/**
 * @class SimClusterMerger
 * @brief Class responsible for merging simulated clusters.
 *
 * The SimClusterMerger class is used to merge simulated clusters based on certain criteria.
 * It provides methods to downscale muons, merge clusters, and calculate the reconstructed energy of a cluster.
 * The class also includes utility functions for sorting and applying sorting in place.
 */

#ifndef SimClusterMerger_h
#define SimClusterMerger_h

#include <algorithm>
#include <cstddef>
#include <numeric>
#include <memory>

#include "Hits.h"
#include "SimClusterMergeWrapper.h"
#include "SimHistoryTool.h"

class SimClusterMerger {
public:
  SimClusterMerger( // const std::vector<Hit>& hits,
      const HitsSoA* hits,
      // find replacement for RecHitTools,
      const SimHistoryTool *history);

  std::vector<SimCluster> downScaleMuons(const std::vector<SimCluster> &scs) const;

  //make overload merge function for neutral stage because it does not 
  //need a unused_clusters vector
  std::vector<SimCluster> merge(std::vector<SimCluster> &scs,
                                const std::vector<SimTrack> &simtracks,
                                std::vector<SimCluster> &unused_clusters,
                                std::vector<std::vector<size_t>> &idxs) const;

  std::vector<SimCluster> merge(std::vector<SimCluster> &scs,
                                std::vector<std::vector<size_t>> &idxs) const;

  void setCNLayers(int cNLayers) { cNLayers_ = cNLayers; }
  void setHighEfracThreshold(float thresh) { isHighEfracThreshold_ = thresh; }
  void setConnectThreshold(float thresh) { connectThreshold_ = thresh; }
  const HitsSoA *hits() const { return hits_; }
  float recEnergy(const SimCluster &cluster) const;

  template <typename T>
  std::vector<std::size_t> argsort(const std::vector<T> &vec);
  template <typename T>
  void apply_argsort_in_place(std::vector<T> &vec,
                              const std::vector<std::size_t> &p) const;

private:
  SimClusterMerger() {}

  double calcCircle(const SimCluster *sc, float &assignedEnergySum,
                    float &regionEnergySum, float &firstHitRadius) const;
  void createHitMap();
  const Hit *getHit(uint32_t) const;

  inline Vector3D getHitPosVec(Hit hit) const {
    return Vector3D(hit.position().x, hit.position().y, hit.position().z);
  }

  float getHitRadius(uint32_t id) const;

  // const std::vector<Hit> * hits_;
  const HitsSoA* hits_;
  // find replacement for RecHitTools
  const SimHistoryTool *histtool_;
  std::map<uint32_t, size_t> hitmap_;

  int cNLayers_;
  float isHighEfracThreshold_;
  float connectThreshold_;

};

template <typename T>
std::vector<std::size_t> SimClusterMerger::argsort(const std::vector<T> &vec) {
  std::vector<size_t> p(vec.size());
  std::iota(p.begin(), p.end(), 0);
  std::sort(p.begin(), p.end(),
            [&](std::size_t i, std::size_t j) { return vec[i] < vec[j]; });
  return p;
}

template <typename T>
void SimClusterMerger::apply_argsort_in_place(
    std::vector<T> &vec, const std::vector<std::size_t> &p) const {
  std::vector<bool> done(vec.size());
  for (std::size_t i = 0; i < vec.size(); ++i) {
    if (done[i]) {
      continue;
    }
    done[i] = true;
    std::size_t prev_j = i;
    std::size_t j = p[i];
    while (i != j) {
      std::swap(vec[prev_j], vec[j]);
      done[j] = true;
      prev_j = j;
      j = p[j];
    }
  }
}

#endif
