#ifndef SimTrack_h
#define SimTrack_h

#include "Tools.h"

/**
 * @class SimTrack
 * @brief Represents a simulated track in a particle physics event.
 *
 * The SimTrack class stores information about a simulated track, including its particle ID, momentum, charge,
 * vertex index, and generator particle index. It also provides methods to access and modify the track's position
 * and momentum at the tracker surface, as well as information about whether the track crossed a boundary and the
 * position and momentum at the boundary.
 */

class SimTrack {
public:
  SimTrack();
  SimTrack(int ipart, int charge, const Vector4D &p, int iv, int ig, const Vector3D &tkp,
           const Vector4D &tkm);

  int vertIndex() const { return ivert_; }
  bool noVertex() const { return ivert_ == -1; }

  int genpartIndex() const { return igenpart_; }
  bool noGenpart() const { return igenpart_ == -1; }

  const Vector3D &trackerSurfacePosition() const { return tkposition_; }
  const Vector4D &trackerSurfaceMomentum() const { return tkmomentum_; }

  inline void setTkPosition(const Vector3D &pos) { tkposition_ = pos; }
  inline void setTkMomentum(const Vector4D &mom) { tkmomentum_ = mom; }
  inline void setVertIndex(const int v) { ivert_ = v; }

  void setCrossedBoundaryVars(bool crossedBoundary, int idAtBoundary,
                              Vector4D positionAtBoundary,
                              Vector4D momentumAtBoundary) {
    crossedBoundary_ = crossedBoundary;
    idAtBoundary_ = idAtBoundary;
    positionAtBoundary_ = positionAtBoundary;
    momentumAtBoundary_ = momentumAtBoundary;
  }
  bool crossedBoundary() const { return crossedBoundary_; }
  const Vector4D &getPositionAtBoundary() const { return positionAtBoundary_; }
  Vector4D getMomentumAtBoundary() const { return momentumAtBoundary_; }
  int getIDAtBoundary() const { return idAtBoundary_; }

  Vector4D momentum() const { return theMomentum_; }
  int type() const { return thePID_; }
  float charge() const { return charge_; };
  void setCharge(float c) { charge_ = c; }

  void setEventId(unsigned int e) { event = e; }
  unsigned int eventId() const { return event; }

  void setTrackId(unsigned int t) { tId_ = t; }
  unsigned int trackId() const { return tId_; }

private:
  unsigned int event;
  unsigned int tId_;
  int thePID_;
  int charge_;

  Vector4D theMomentum_;

  int ivert_;
  int igenpart_;

  Vector3D tkposition_;
  Vector4D tkmomentum_;

  bool crossedBoundary_;
  int idAtBoundary_;
  Vector4D positionAtBoundary_;
  Vector4D momentumAtBoundary_;
};

#endif
