#include "../interface/DataLoader.h"
#include "TTree.h"
#include "TTreeReader.h"
#include "TTreeReaderValue.h"
#include "TTreeReaderArray.h"
#include <algorithm>
#include <cstdint>
#include <sys/types.h>

DataLoader::DataLoader(std::vector<std::string> files) : input_files_v_(files) {
  std::vector<std::string> tree_names = {"pfTruthDumper/pfParticles",
                                         "pfTruthDumper/hits", 
                                         "pfTruthDumper/simClusters",
                                         "pfTruthDumper/simTracks",
                                         "pfTruthDumper/simVertices",
                                         "pfTruthDumper/simHits"};
  std::string root_file = "pfTruthDumper.root";

  auto snapopts = ROOT::RDF::RSnapshotOptions();
  snapopts.fMode = "UPDATE";

  for (unsigned int i = 0; i < input_files_v_.size(); ++i) {      
    loadFromCSV(input_files_v_[i], tree_names[i], root_file, snapopts);
  }

  input_file_ = std::make_shared<TFile>("pfTruthDumper.root");

  fillSimHitsSoA();
  fillHitsSoA();
  fillSimClusters();
  fillSimVertices();
  fillSimTracks();
  fillPFParticles();
}

DataLoader::DataLoader(std::shared_ptr<TFile> file) : input_file_(file), input_files_v_({"0"}) {
  fillHitsSoA();
  fillSimClusters();
  fillSimVertices();
  fillSimTracks();
}

template <typename T>
std::vector<std::vector<T>> DataLoader::fillVector(const char *treeName,
                                                   const char *branchName,
                                                   std::shared_ptr<TFile> file,
                                                   bool isFlatDataFrame) {
  TTreeReader reader(treeName, file.get());
  std::vector<std::vector<T>> result;
  
  if (isFlatDataFrame) {
    TTreeReaderValue<double> reader_value(reader, branchName);
    std::vector<T> tmp_result;
    while (reader.Next()) {
      tmp_result.push_back(*reader_value);
    }
    result.push_back(tmp_result);
  } else {
    TTreeReaderValue<std::vector<T>> reader_value(reader, branchName);
    while (reader.Next()) {
      result.push_back(*reader_value);
    }
  }
  return result;
}

void DataLoader::loadFromCSV(std::string csv_file, 
                             std::string treeName, 
                             std::string root_file, 
                             ROOT::RDF::RSnapshotOptions snapopts) { 
  auto rdf = ROOT::RDF::FromCSV(csv_file, true, ',', -1LL);
  rdf.Snapshot(treeName,  root_file, "", snapopts);
}

void DataLoader::fillPFParticles() {
  std::vector<std::vector<float>> pfParticleCharge = fillVector<float>(
      "pfTruthDumper/pfParticles", "pfParticle_charge", input_file_, true);
  std::vector<std::vector<int>> pfParticlePdgId = fillVector<int>(
      "pfTruthDumper/pfParticles", "pfParticle_pdgId", input_file_, true);
  /*std::vector<std::vector<std::vector<int>>> pfParticleSimClusters =
      fillVector<std::vector<int>>("pfTruthDumper/pfParticles",
                                   "pfParticle_simClusterIndex", input_file_, true); 
  std::vector<std::vector<std::vector<int>>> pfParticleSimTracks = 
      fillVector<std::vector<int>>("pfTruthDumper/pfParticles",
                                   "pfParticle_simTrackIndex", input_file_, true);
  */
  unsigned int nEvents = pfParticleCharge.size();
  pfParticles_.resize(nEvents);
  std::vector<std::vector<int>> pfParticleSimClusters, pfParticleSimTracks;
  pfParticleSimClusters.resize(nEvents);
  pfParticleSimTracks.resize(nEvents);
  for (unsigned int i = 0; i < nEvents; ++i) {
    unsigned int nPFParticles = pfParticleCharge[i].size();
    for (unsigned int j = 0; j < nPFParticles; ++j) {
      std::vector<int> pfParticleSimClusters, pfParticleSimTracks;
      for (unsigned int l = 0; l < simClusters_[i].size(); ++l) {
        if (simClusters_[i][l].genParticle() == static_cast<int>(j)) {
          pfParticleSimClusters.push_back(l);
        }
      }
      for (auto& track : simTracks_[i]) {
        if (track.genpartIndex() == static_cast<int>(j))
          pfParticleSimTracks.push_back(static_cast<int>(j));
      }
      PFParticle pfParticle(pfParticleSimClusters,
                            pfParticleSimTracks,
                            pfParticleCharge[i][j], 
                            pfParticlePdgId[i][j]);
      pfParticles_[i].push_back(pfParticle);
    } 
  }
}




void DataLoader::fillSimHitsSoA() {
  std::vector<std::vector<float>> simHitsEnergy =
      fillVector<float>("pfTruthDumper/simHits", "simHit_energy", input_file_, true);
  std::vector<std::vector<float>> simHitsPos_x = 
      fillVector<float>("pfTruthDumper/simHits", "simHit_x", input_file_, true);
  std::vector<std::vector<float>> simHitsPos_y =
      fillVector<float>("pfTruthDumper/simHits", "simHit_y", input_file_, true);
  std::vector<std::vector<float>> simHitsPos_z =
      fillVector<float>("pfTruthDumper/simHits", "simHit_z", input_file_, true);
  std::vector<std::vector<float>> simHitsTime =
      fillVector<float>("pfTruthDumper/simHits", "simHit_t", input_file_, true);
  std::vector<std::vector<float>> simHitsFraction =
      fillVector<float>("pfTruthDumper/simHits", "simHit_fraction", input_file_, true);
  std::vector<std::vector<uint32_t>> simHitsId = 
      fillVector<uint32_t>("pfTruthDumper/simHits", "simHit_detid", input_file_, true);
  std::vector<std::vector<int>> simHitsClusterIndex = 
      fillVector<int>("pfTruthDumper/simHits", "simHit_clusterIndex", input_file_, true);

  unsigned int nEvents = simHitsId.size();
  simHitsSoA_.clear();
  simHitsSoA_.resize(nEvents);
  for (unsigned int i = 0; i < nEvents; ++i) {
    unsigned int nHits = simHitsId[i].size();
    simHitsSoA_[i].resize(nHits);
    for (unsigned int j = 0; j < nHits; ++j) {
      simHitsSoA_[i].energy[j] = simHitsEnergy[i][j];
      simHitsSoA_[i].x[j] = simHitsPos_x[i][j];
      simHitsSoA_[i].y[j] = simHitsPos_y[i][j];
      simHitsSoA_[i].z[j] = simHitsPos_z[i][j];
      simHitsSoA_[i].time[j] = simHitsTime[i][j];
      simHitsSoA_[i].fraction[j] = simHitsFraction[i][j];
      simHitsSoA_[i].id[j] = simHitsId[i][j];
      simHitsSoA_[i].clusterIndex[j] = simHitsClusterIndex[i][j];
    }
  }
}


void DataLoader::fillHitsSoA() {
  std::vector<std::vector<float>> hitsEnergy =
      fillVector<float>("pfTruthDumper/hits", "hits_energy", input_file_, true);
  std::vector<std::vector<float>> hitsPos_x =
      fillVector<float>("pfTruthDumper/hits", "hits_x", input_file_, true);
  std::vector<std::vector<float>> hitsPos_y =
      fillVector<float>("pfTruthDumper/hits", "hits_y", input_file_, true);
  std::vector<std::vector<float>> hitsPos_z =
      fillVector<float>("pfTruthDumper/hits", "hits_z", input_file_, true);
  std::vector<std::vector<float>> hitsTime =
      fillVector<float>("pfTruthDumper/hits", "hits_t", input_file_, true);
  std::vector<std::vector<uint32_t>> hitsId =
      fillVector<uint32_t>("pfTruthDumper/hits", "hits_id", input_file_, true);
  unsigned int nEvents = hitsId.size(); 
  hitsSoA_.clear();
  hitsSoA_.resize(nEvents);
  for (unsigned int i = 0; i < nEvents; ++i) {
    unsigned int nHits = hitsId[i].size();
    hitsSoA_[i].resize(nHits);
    for (unsigned int j = 0; j < nHits; ++j) {
      hitsSoA_[i].energy[j] = hitsEnergy[i][j];
      hitsSoA_[i].x[j] = hitsPos_x[i][j];
      hitsSoA_[i].y[j] = hitsPos_y[i][j];
      hitsSoA_[i].z[j] = hitsPos_z[i][j];
      hitsSoA_[i].time[j] = hitsTime[i][j];
      hitsSoA_[i].fraction[j] = 0.f;
      hitsSoA_[i].id[j] = hitsId[i][j];
    }
  }
}

void DataLoader::fillSimClusters() {
  std::vector<std::vector<float>> simClusterEnergy = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_energy", input_file_, true);
  std::vector<std::vector<float>> simClusterWidth_x = fillVector<float>(
      "pfTruthDumper/simClusters","simCluster_width_x", input_file_, true);
  std::vector<std::vector<float>> simClusterWidth_y = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_width_y", input_file_, true);
  std::vector<std::vector<float>> simClusterWidth_z = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_width_z", input_file_, true);
  std::vector<std::vector<float>> impactMomentum_x = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_impactMomentum_x", input_file_, true);
  std::vector<std::vector<float>> impactMomentum_y = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_impactMomentum_y", input_file_, true);
  std::vector<std::vector<float>> impactMomentum_z = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_impactMomentum_z", input_file_, true);
  std::vector<std::vector<float>> impactMomentum_t = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_impactMomentum_t", input_file_, true);
  std::vector<std::vector<float>> posBoundary_x = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_posBoundary_x", input_file_, true);
  std::vector<std::vector<float>> posBoundary_y = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_posBoundary_y", input_file_, true);
  std::vector<std::vector<float>> posBoundary_z = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_posBoundary_z", input_file_, true);
  std::vector<std::vector<float>> impactPoint_x = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_impactPoint_x", input_file_, true);
  std::vector<std::vector<float>> impactPoint_y = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_impactPoint_y", input_file_, true);
  std::vector<std::vector<float>> impactPoint_z = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_impactPoint_z", input_file_, true);
  std::vector<std::vector<float>> impactPoint_t = fillVector<float>(
      "pfTruthDumper/simClusters", "simCluster_impactPoint_t", input_file_, true);
  std::vector<std::vector<int>> simCluster_pdgId = fillVector<int>(
      "pfTruthDumper/simClusters", "simCluster_pdgId", input_file_, true);
  std::vector<std::vector<int>> simCluster_layer = fillVector<int>(
      "pfTruthDumper/simClusters", "simCluster_layer", input_file_, true);
  std::vector<std::vector<int>> simCluster_pfParticle = fillVector<int>(
      "pfTruthDumper/simClusters", "simCluster_genParticle", input_file_, true);
  /*std::vector<hits_in_sc> simCluster_hits =
      fillVector<std::vector<uint32_t>>("pfTruthDumper/simClusters",
                                        "simCluster_hits", input_file_, true);
  std::vector<fractions_in_sc> simCluster_fractions =
      fillVector<std::vector<float>>("pfTruthDumper/simClusters",
                                     "simCluster_fractions", input_file_, true);*/

  unsigned int nEvents =
      impactMomentum_x.size(); 
  simClusters_.resize(nEvents);
  for (unsigned int i = 0; i < nEvents; ++i) {
    unsigned int nSimClusters = impactMomentum_x[i].size();
    auto nSimHits = simHitsSoA_[i].id.size();
    for (unsigned int j = 0; j < nSimClusters; ++j) {
      std::vector<uint32_t> hits;
      std::vector<float> fractions;
      for (unsigned int k = 0; k < nSimHits; ++k) {
        if (simHitsSoA_[i].clusterIndex[k] == static_cast<int>(j)) {
          hits.push_back(simHitsSoA_[i].id[k]);
          fractions.push_back(simHitsSoA_[i].fraction[k]);
        }
      }
      Vector4D impactMomentum(impactMomentum_x[i][j], impactMomentum_y[i][j],
                              impactMomentum_z[i][j], impactMomentum_t[i][j]);
      Vector3D positionAtBoundary(posBoundary_x[i][j], posBoundary_y[i][j],
                                  posBoundary_z[i][j]);
      Vector3D impactPoint(impactPoint_x[i][j], impactPoint_y[i][j],
                           impactPoint_z[i][j]);
      SimCluster simCl(impactMomentum, positionAtBoundary, impactPoint,
                       simClusterEnergy[i][j],
                       simClusterWidth_x[i][j],
                       simClusterWidth_y[i][j],
                       simClusterWidth_z[i][j], 
                       simCluster_layer[i][j],
                       //simCluster_hits[i][j], simCluster_fractions[i][j],
                       hits, fractions,
                       simCluster_pdgId[i][j],
                       simCluster_pfParticle[i][j]);
      simClusters_[i].push_back(simCl);
    }
  }
}

void DataLoader::fillSimVertices() {
  std::vector<std::vector<float>> vertex_x = fillVector<float>(
      "pfTruthDumper/simVertices", "simVertex_x", input_file_, true);
  std::vector<std::vector<float>> vertex_y = fillVector<float>(
      "pfTruthDumper/simVertices", "simVertex_y", input_file_, true);
  std::vector<std::vector<float>> vertex_z = fillVector<float>(
      "pfTruthDumper/simVertices", "simVertex_z", input_file_, true);
  std::vector<std::vector<float>> vertex_t = fillVector<float>(
      "pfTruthDumper/simVertices", "simVertex_t", input_file_, true);
  std::vector<std::vector<int>> vertex_parentIndex = fillVector<int>(
      "pfTruthDumper/simVertices", "simVertex_parentIndex", input_file_, true);
  std::vector<std::vector<int>> vertex_vertexIndex = fillVector<int>(
      "pfTruthDumper/simVertices", "simVertex_vertexIndex", input_file_, true);
  unsigned int nEvents = vertex_x.size(); 
  simVertices_.resize(nEvents);
  for (unsigned int i = 0; i < nEvents; ++i) {
    unsigned int nSimVertices = vertex_x[i].size();
    for (unsigned int j = 0; j < nSimVertices; ++j) {
      Vector3D vertex(vertex_x[i][j], vertex_y[i][j], vertex_z[i][j]);
      float tof = vertex_t[i][j];
      int parentIndex = vertex_parentIndex[i][j];
      unsigned int vertexIndex = vertex_vertexIndex[i][j];
      SimVertex simVtx(vertex, tof, parentIndex, vertexIndex);
      simVertices_[i].push_back(simVtx);
    }
  }
}

void DataLoader::fillSimTracks() {
  std::vector<std::vector<float>> simtrack_momentum_x = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_momentum_x", input_file_, true);
  std::vector<std::vector<float>> simtrack_momentum_y = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_momentum_y", input_file_, true);
  std::vector<std::vector<float>> simtrack_momentum_z = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_momentum_z", input_file_, true);
  std::vector<std::vector<float>> simtrack_momentum_t = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_momentum_t", input_file_, true);
  std::vector<std::vector<float>> simtrack_tkx = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_track_x", input_file_, true);
  std::vector<std::vector<float>> simtrack_tky = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_track_y", input_file_, true);
  std::vector<std::vector<float>> simtrack_tkz = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_track_z", input_file_, true);
  std::vector<std::vector<float>> simtrack_tkmx = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_trackmomentum_x", input_file_, true);
  std::vector<std::vector<float>> simtrack_tkmy = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_trackmomentum_y", input_file_, true);
  std::vector<std::vector<float>> simtrack_tkmz = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_trackmomentum_z", input_file_, true);
  std::vector<std::vector<float>> simtrack_tkmt = fillVector<float>(
      "pfTruthDumper/simTracks", "simTrack_trackmomentum_t", input_file_, true);
  std::vector<std::vector<int>> simtrack_type = fillVector<int>(
      "pfTruthDumper/simTracks", "simTrack_tkType", input_file_, true);
  std::vector<std::vector<int>> simtrack_vertex = fillVector<int>(
      "pfTruthDumper/simTracks", "simTrack_vertexIndex", input_file_, true);
  std::vector<std::vector<int>> simtrack_genp = fillVector<int>(
      "pfTruthDumper/simTracks", "simTrack_genParticle", input_file_, true);
  std::vector<std::vector<int>> simtrack_charge = fillVector<int>(
      "pfTruthDumper/simTracks", "simTrack_charge", input_file_, true);
  unsigned int nEvents = simtrack_tkx.size();
  simTracks_.resize(nEvents);
  for (unsigned int i = 0; i < nEvents; ++i) {
    unsigned int nSimTracks = simtrack_tkx[i].size();
    for (unsigned int j = 0; j < nSimTracks; ++j) {
      Vector4D momentum(simtrack_momentum_x[i][j], simtrack_momentum_y[i][j],
                        simtrack_momentum_z[i][j], simtrack_momentum_t[i][j]);
      int type = simtrack_type[i][j];
      int vertex = simtrack_vertex[i][j];
      int genparticle = simtrack_genp[i][j];
      int charge = simtrack_charge[i][j];
      Vector3D trackPosition(simtrack_tkx[i][j], simtrack_tky[i][j],
                             simtrack_tkz[i][j]);
      Vector4D trackMomentum(simtrack_tkmx[i][j], simtrack_tkmy[i][j],
                             simtrack_tkmz[i][j], simtrack_tkmt[i][j]);
      SimTrack simTk(type, charge, momentum, vertex, genparticle, trackPosition,
                     trackMomentum);
      simTracks_[i].push_back(simTk);
    }
  }
}

void DataLoader::writeMergedSimClusters(std::vector<std::vector<SimCluster>>& merged_scs) {
  std::vector<std::vector<float>> posBoundary_x, posBoundary_y, posBoundary_z, 
                                  impactPoint_x, impactPoint_y, impactPoint_z,
                                  impactMomentum_x, impactMomentum_y, impactMomentum_z, impactMomentum_t,
                                  energy;
  std::vector<std::vector<uint32_t>> event;
  std::vector<std::vector<int>> pdgId, layer;
  std::vector<fractions_in_sc> fractions;
  std::vector<hits_in_sc> hits;

  std::shared_ptr<TFile> output(TFile::Open("mergedClusters.root", "RECREATE"));
  TTree* merged_scs_tree = new TTree ("mergedSimClusters", "mergedSimClusters");
  merged_scs_tree->Branch("mergedSimCluster_posBoundary_x", &posBoundary_x);
  merged_scs_tree->Branch("mergedSimCluster_posBoundary_y", &posBoundary_y);
  merged_scs_tree->Branch("mergedSimCluster_posBoundary_z", &posBoundary_z);
  merged_scs_tree->Branch("mergedSimCluster_impactPoint_x", &impactPoint_x);
  merged_scs_tree->Branch("mergedSimCluster_impactPoint_y", &impactPoint_y);
  merged_scs_tree->Branch("mergedSimCluster_impactPoint_z", &impactPoint_z);
  merged_scs_tree->Branch("mergedSimCluster_impactMomentum_x", &impactMomentum_x);
  merged_scs_tree->Branch("mergedSimCluster_impactMomentum_y", &impactMomentum_y);
  merged_scs_tree->Branch("mergedSimCluster_impactMomentum_z", &impactMomentum_z);
  merged_scs_tree->Branch("mergedSimCluster_impactMomentum_t", &impactMomentum_t);
  merged_scs_tree->Branch("mergedSimCluster_energy", &energy);
  merged_scs_tree->Branch("mergedSimCluster_event", &event);
  merged_scs_tree->Branch("mergedSimCluster_layer", &layer);
  merged_scs_tree->Branch("mergedSimCluster_pdgId", &pdgId);
  merged_scs_tree->Branch("mergedSimCluster_hits", &hits);
  merged_scs_tree->Branch("mergedSimCluster_fractions", &fractions);

  unsigned int nEvents = merged_scs.size();
  posBoundary_x.resize(nEvents);
  posBoundary_y.resize(nEvents);
  posBoundary_z.resize(nEvents);
  impactPoint_x.resize(nEvents);
  impactPoint_y.resize(nEvents);
  impactPoint_z.resize(nEvents);
  impactMomentum_x.resize(nEvents);
  impactMomentum_y.resize(nEvents);
  impactMomentum_z.resize(nEvents);
  impactMomentum_t.resize(nEvents);
  energy.resize(nEvents);
  event.resize(nEvents);
  layer.resize(nEvents);
  pdgId.resize(nEvents);
  hits.resize(nEvents);
  fractions.resize(nEvents);

  for (unsigned int i = 0; i < nEvents; ++i) {
    unsigned int nMergedSimClusters = merged_scs[i].size();
    hits[i].resize(nMergedSimClusters);
    fractions[i].resize(nMergedSimClusters);
    for (unsigned int j = 0; j < nMergedSimClusters; ++j) {
      posBoundary_x[i].push_back(merged_scs[i][j].positionAtBoundary().x);
      posBoundary_y[i].push_back(merged_scs[i][j].positionAtBoundary().y);
      posBoundary_z[i].push_back(merged_scs[i][j].positionAtBoundary().z);
      impactPoint_x[i].push_back(merged_scs[i][j].impactPoint().x);
      impactPoint_y[i].push_back(merged_scs[i][j].impactPoint().y);
      impactPoint_z[i].push_back(merged_scs[i][j].impactPoint().z);
      impactMomentum_x[i].push_back(merged_scs[i][j].impactMomentum().x1);
      impactMomentum_y[i].push_back(merged_scs[i][j].impactMomentum().x2);
      impactMomentum_z[i].push_back(merged_scs[i][j].impactMomentum().x3);
      impactMomentum_t[i].push_back(merged_scs[i][j].impactMomentum().y);
      energy[i].push_back(merged_scs[i][j].energy());
      event[i].push_back(i);
      layer[i].push_back(merged_scs[i][j].getLayer());
      pdgId[i].push_back(merged_scs[i][j].pdgId());

      auto haf = merged_scs[i][j].hits_and_fractions();
      hits[i][j].resize(haf.size());
      fractions[i][j].resize(haf.size());
      for (const auto& h : haf) {
        hits[i][j].push_back(h.first);
        fractions[i][j].push_back(h.second);
      }
    }
  }

  merged_scs_tree->Fill();
  merged_scs_tree->Write();
}
