#include "../interface/HitMergeVertex.h"
#include <limits>

float HitMergeVertex::connectThreshold = 0.2;
float HitMergeVertex::singleClusterEfrac = 0.9;

HitMergeVertex::HitMergeVertex(
    SimClusterMergeWrapper* t, const HitsSoA* hits,
    const std::vector<std::pair<uint32_t, float>> &hitsAndFractions)
    : MergeVertex<SimClusterMergeWrapper>(t) {
  for (size_t i = 0; i < hitsAndFractions.size(); ++i) {
    auto id = hitsAndFractions.at(i).first;
    auto nHits = hits->id.size();
    for (unsigned j = 0; j < nHits; ++j) {
      if (hits->id[j] == id) {
        hitsAndFractions_.push_back(
            HitAndFraction(hits->x[j], 
                           hits->y[j], 
                           hits->z[j], 
                           hits->energy[j], 
                           hits->time[j], 
                           hitsAndFractions.at(i).second,
                           hits->id[j]));
        break;
      }
    }
  }
  std::sort(hitsAndFractions_.begin(), hitsAndFractions_.end());
  updateEnergies();
  updateIsLow();
  cached_score_.first = 0;
}

bool HitMergeVertex::shouldConnect(const HitMergeVertex* rhs) const {
  if (this == rhs) 
    return false;
  float s = calcScore(rhs);
  return s > connectThreshold &&
         MergeVertex<SimClusterMergeWrapper>::shouldConnect(rhs);
}

void HitMergeVertex::updateIsLow() {
  float fractions_sum = 0, energy_sum = 0;
  for (const auto &h : hitsAndFractions_) {
    fractions_sum += h.fraction * (h.fraction * h.energy);
    energy_sum += h.fraction * h.energy;
  }
  float energy_fraction = fractions_sum / (energy_sum + 1e-6);
  isLow_ = energy_fraction < singleClusterEfrac;
}

void HitMergeVertex::absorb(HitMergeVertex* rhs) {
  absorbHitsAndFractions(rhs->hitsAndFractions_);
  rhs->hitsAndFractions_.clear();

  updateEnergies();
  rhs->updateEnergies();

  MergeVertex<SimClusterMergeWrapper>::absorb(rhs);
}

float HitMergeVertex::calcScore(const HitMergeVertex* rhs) const {
  if (!rhs->valid())
    return 0;
  if (cached_score_.first == rhs) {
    cached_score_.first = 0;
    return cached_score_.second;
  }
  std::vector<size_t> int_a, int_b;
  intersectSorted(hitsAndFractions_, rhs->hitsAndFractions_, int_a, int_b);
  float score = 0;
  for (size_t i = 0; i < int_a.size(); ++i) {
    const auto &h = hitsAndFractions_.at(int_a.at(i));
    const auto &h_rhs = rhs->hitsAndFractions_.at(int_b.at(i));
    float uE = h.energy * (h.fraction + h_rhs.fraction);
    score += uE;
  }
  score /= rhs->energy_ + 1e-6;
  cached_score_.first = rhs;
  cached_score_.second = score;
  
  return score;
}

void HitMergeVertex::updateEnergies() {
  energy_ = 0;
  energyAllHits_ = 0;
  for (const auto &h : hitsAndFractions_) {
    energyAllHits_ += h.energy;
    energy_ += h.energy * h.fraction;
  }
}

void HitMergeVertex::absorbHitsAndFractions(
    const std::vector<HitAndFraction> &h_rhs) {
  std::vector<size_t> sidx, rhs_sidx, _uidx, rhs_uidx;

  intersectSorted(hitsAndFractions_, h_rhs, sidx, rhs_sidx, _uidx, rhs_uidx);

  for (size_t i = 0; i < sidx.size(); ++i) {
    hitsAndFractions_.at(sidx.at(i)).fraction +=
        h_rhs.at(rhs_sidx.at(i)).fraction;
  }

  std::vector<HitAndFraction> rhsu;
  for (size_t i = 0; i < rhs_uidx.size(); ++i) {
    rhsu.push_back(h_rhs.at(rhs_uidx.at(i)));
  }

  hitsAndFractions_.insert(hitsAndFractions_.end(), rhsu.begin(), rhsu.end());
  std::sort(hitsAndFractions_.begin(), hitsAndFractions_.end());

  return;
  insertSortedInSorted(rhsu, hitsAndFractions_);
}
