#include "../interface/SimClusterMerger.h"
#include "../interface/MergeGraph.h"
#include <cmath>
#include <cstddef>
#include <limits>

SimClusterMerger::SimClusterMerger(const HitsSoA* hits,
                                   const SimHistoryTool *hist)
    : hits_(hits), histtool_(hist) {
  createHitMap();
  cNLayers_ = 10;
  isHighEfracThreshold_ = 0.85; 
  connectThreshold_ = 0.3;
}

std::vector<SimCluster> SimClusterMerger::downScaleMuons(
    const std::vector<SimCluster> &scs) const {
  std::vector<SimCluster> out;
  for (const auto &sc : scs) {
    if (abs(sc.pdgId()) == 13) {
      auto sc2 = sc;
      float recenergy = recEnergy(sc2);
      auto sc2mom = sc2.impactMomentum();
      float momMag = sc2mom.P();
      float rescaleFactor = recenergy / momMag;
      sc2.setImpactMomentum(Vector4D(sc2mom.x1 / rescaleFactor,
                                     sc2mom.x2 / rescaleFactor,
                                     sc2mom.x3 / rescaleFactor, sc2mom.y));
      out.push_back(sc2);
    } else {
      out.push_back(sc);
    }
  }
  return out;
}

std::vector<SimCluster> SimClusterMerger::merge(std::vector<SimCluster> &scs,
                                                const std::vector<SimTrack> &simtracks,
                                                std::vector<SimCluster> &unused_clusters,
                                                std::vector<std::vector<size_t>> &idxs) const {

  HitMergeVertex::connectThreshold = connectThreshold_;
  HitMergeVertex::singleClusterEfrac = isHighEfracThreshold_;

  std::vector<SimClusterMergeWrapper> scwrappers;

  size_t idx = 0;
  for (auto &sc : scs) {
    scwrappers.push_back(SimClusterMergeWrapper(&sc, idx));
    idx++;
  }
  MergeGraph<HitMergeVertex> graph;

  //for (auto &scw : scwrappers) {
  for (size_t i = 0; i < scwrappers.size(); ++i) {
    auto hafs_in = (scwrappers[i].simCluster())->hits_and_fractions();
    std::vector<std::pair<uint32_t, float>> hafs;
    for (const auto &hf : hafs_in) {
      if (scwrappers[i].getLayer() > (size_t)cNLayers_)
        continue;
      hafs.push_back(hf);
    }
    HitMergeVertex vertex(&scwrappers[i], hits_, hafs);
    SimCluster *sc = scwrappers[i].simCluster();
    int genParticle = sc->genParticle();
    auto simTrackIt = std::find_if(simtracks.begin(), simtracks.end(), 
                                [&](const SimTrack &st) { return st.genpartIndex() == genParticle; });
    // check compatibility between simTrack momentum and simCluster energy
    // we assume massless particles
    const float simTrack_momentum = simTrackIt->momentum().P();
    const float simCluster_energy = sc->energy();
    if (abs(simTrack_momentum - simCluster_energy) / simCluster_energy < 0.3) {
      unused_clusters.push_back(*sc);
    }
    graph.addVertex(HitMergeVertex(&scwrappers[i], hits_, hafs));
  }
 
  graph.createEdges();
  graph.merge();
  std::vector<SimCluster> out;
  idxs.clear();
  auto result = graph.vertices();
  
  for (const auto &vert : result) {
    if (!vert.valid())
      continue;
    std::vector<size_t> tmpidx;
    std::vector<SimCluster *> tomerge;
    for (const auto &scw : vert.objects()) {
      tmpidx.push_back(scw->idx());
      tomerge.push_back(scw->simCluster());
    }
    out.push_back(histtool_->createMergedSimCluster(tomerge));
    idxs.push_back(tmpidx);
  }
  idxs.shrink_to_fit();
  return out;
}

std::vector<SimCluster> SimClusterMerger::merge(std::vector<SimCluster> &scs,
                                                std::vector<std::vector<size_t>> &idxs) const {

  HitMergeVertex::connectThreshold = connectThreshold_;
  HitMergeVertex::singleClusterEfrac = isHighEfracThreshold_;

  std::vector<SimClusterMergeWrapper> scwrappers;

  size_t idx = 0;
  for (auto &sc : scs) {
    scwrappers.push_back(SimClusterMergeWrapper(&sc, idx));
    idx++;
  }
  MergeGraph<HitMergeVertex> graph;

  //for (auto &scw : scwrappers) {
  for (size_t i = 0; i < scwrappers.size(); ++i) {
    auto hafs_in = (scwrappers[i].simCluster())->hits_and_fractions();
    std::vector<std::pair<uint32_t, float>> hafs;
    for (const auto &hf : hafs_in) {
      if (scwrappers[i].getLayer() > (size_t)cNLayers_)
        continue;
      hafs.push_back(hf);
    }
    HitMergeVertex vertex(&scwrappers[i], hits_, hafs);
    graph.addVertex(HitMergeVertex(&scwrappers[i], hits_, hafs));
  }
 
  graph.createEdges();
  graph.merge();
  std::vector<SimCluster> out;
  idxs.clear();
  auto result = graph.vertices();
  
  for (const auto &vert : result) {
    if (!vert.valid())
      continue;
    std::vector<size_t> tmpidx;
    std::vector<SimCluster *> tomerge;
    for (const auto &scw : vert.objects()) {
      tmpidx.push_back(scw->idx());
      tomerge.push_back(scw->simCluster());
    }
    out.push_back(histtool_->createMergedSimCluster(tomerge));
    idxs.push_back(tmpidx);
  }
  idxs.shrink_to_fit();
  return out;
}


float SimClusterMerger::recEnergy(const SimCluster &cluster) const {
  float en = 0;
  auto hits_and_fractions = cluster.hits_and_fractions();
  unsigned int nHits = hits_and_fractions.size();
  for (unsigned int i = 0; i < nHits; ++i) {
    auto findHit = std::find(hits_->id.begin(), hits_->id.end(),
                             hits_and_fractions[i].first);
    if (findHit == hits_->id.end())
      continue;
    auto index = std::distance(hits_->id.begin(), findHit);
    en += hits_->energy[index] * hits_and_fractions[i].second;
  }
  return en;
}

void SimClusterMerger::createHitMap() {
  hitmap_.clear();
  for (size_t i = 0; i < hits_->id.size(); ++i) {
    hitmap_[hits_->id[i]] = i;
  }
}
